/**
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Jianfeng Gao ( jianfeng dot gao at kit dot edu )
 * @date       2022
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

#include <ArmarXCore/core/logging/Logging.h>
#include <ArmarXCore/interface/serialization/Eigen.h>
#include <ArmarXCore/util/CPPUtility/TripleBuffer.h>
#include "MP.h"


namespace armarx::control::common::mp
{
    struct JSMPInput : virtual MPInput
    {
        Eigen::VectorXf angleRadian;
        Eigen::VectorXf angularVel;
        double deltaT;
    };

    struct JSMPOutput : virtual MPOutput
    {
        Eigen::VectorXf angleRadian;
        Eigen::VectorXf angularVel;
    };

    using JSMPInputPtr  = std::shared_ptr<JSMPInput>;
    using JSMPOutputPtr = std::shared_ptr<JSMPOutput>;



    class JSMP : virtual public MP
    {
    public:
//        JSMP(const std::string& name, nlohmann::json& userConfig, nlohmann::json& defaultConfig);
        JSMP(const MPConfig& c);
        ~JSMP(){};

        void run(MPInputPtr, MPOutputPtr) override;

        Eigen::Quaterniond previousAngularVelocity = Eigen::Quaterniond::Identity();
    };

    typedef std::shared_ptr<JSMP> JSMPPtr;

} /// namespace armarx::control::common::mp
